@@ -3,15 +3,9 @@
 #include "visited_list_pool.h"
 #include "hnswlib.h"
 #include <random>
-#include <iostream>
-#include <fstream>
-#include <stdio.h>
 #include <stdlib.h>
-#include <string.h>
-#include <algorithm>
-#include <atomic>
 #include <unordered_set>
-#include <unordered_map>
+#include <list>
 
 
 // Taken from RcppAnnoy, https://cran.r-project.org/package=RcppAnnoy
@@ -43,7 +37,7 @@ namespace hnswlib {
                 link_list_locks_(max_elements), element_levels_(max_elements) {
             max_elements_ = max_elements;
 
-
+            has_deletions_=false;
             data_size_ = s->get_data_size();
             fstdistfunc_ = s->get_dist_func();
             dist_func_param_ = s->get_dist_func_param();
@@ -120,7 +114,6 @@ namespace hnswlib {
         tableint enterpoint_node_;
 
 
-
         size_t size_links_level0_;
         size_t offsetData_, offsetLevel0_;
 
@@ -129,8 +122,11 @@ namespace hnswlib {
         char **linkLists_;
         std::vector<int> element_levels_;
 
-
         size_t data_size_;
+
+        bool has_deletions_;
+
+
         size_t label_offset_;
         DISTFUNC<dist_t> fstdistfunc_;
         void *dist_func_param_;
@@ -144,6 +140,10 @@ namespace hnswlib {
             return return_label;
         }
 
+        inline void setExternalLabel(tableint internal_id, labeltype label) const {
+            memcpy((data_level0_memory_ + internal_id * size_data_per_element_ + label_offset_), &label, sizeof(labeltype));
+        }
+
         inline labeltype *getExternalLabeLp(tableint internal_id) const {
             return (labeltype *) (data_level0_memory_ + internal_id * size_data_per_element_ + label_offset_);
         }
@@ -159,24 +159,28 @@ namespace hnswlib {
         }
 
         std::priority_queue<std::pair<dist_t, tableint>, std::vector<std::pair<dist_t, tableint>>, CompareByFirst>
-        searchBaseLayer(tableint enterpoint_id, void *data_point, int layer) {
+        searchBaseLayer(tableint ep_id, void *data_point, int layer) {
             VisitedList *vl = visited_list_pool_->getFreeVisitedList();
             vl_type *visited_array = vl->mass;
             vl_type visited_array_tag = vl->curV;
 
             std::priority_queue<std::pair<dist_t, tableint>, std::vector<std::pair<dist_t, tableint>>, CompareByFirst> top_candidates;
             std::priority_queue<std::pair<dist_t, tableint>, std::vector<std::pair<dist_t, tableint>>, CompareByFirst> candidateSet;
-            dist_t dist = fstdistfunc_(data_point, getDataByInternalId(enterpoint_id), dist_func_param_);
 
-            top_candidates.emplace(dist, enterpoint_id);
-            candidateSet.emplace(-dist, enterpoint_id);
-            visited_array[enterpoint_id] = visited_array_tag;
-            dist_t lowerBound = dist;
+            dist_t lowerBound;
+            if (!isMarkedDeleted(ep_id)) {
+                dist_t dist = fstdistfunc_(data_point, getDataByInternalId(ep_id), dist_func_param_);
+                top_candidates.emplace(dist, ep_id);
+                lowerBound = dist;
+                candidateSet.emplace(-dist, ep_id);
+            } else {
+                lowerBound = std::numeric_limits<dist_t>::max();
+                candidateSet.emplace(-lowerBound, ep_id);
+            }
+            visited_array[ep_id] = visited_array_tag;
 
             while (!candidateSet.empty()) {
-
                 std::pair<dist_t, tableint> curr_el_pair = candidateSet.top();
-
                 if ((-curr_el_pair.first) > lowerBound) {
                     break;
                 }
@@ -187,40 +191,47 @@ namespace hnswlib {
                 std::unique_lock <std::mutex> lock(link_list_locks_[curNodeNum]);
 
                 int *data;// = (int *)(linkList0_ + curNodeNum * size_links_per_element0_);
-                if (layer == 0)
-                    data = (int *) (data_level0_memory_ + curNodeNum * size_data_per_element_ + offsetLevel0_);
-                else
-                    data = (int *) (linkLists_[curNodeNum] + (layer - 1) * size_links_per_element_);
-                int size = *data;
+                if (layer == 0) {
+                    data = (int*)get_linklist0(curNodeNum);
+                } else {
+                    data = (int*)get_linklist(curNodeNum, layer);
+//                    data = (int *) (linkLists_[curNodeNum] + (layer - 1) * size_links_per_element_);
+                }
+                size_t size = getListCount((linklistsizeint*)data);
                 tableint *datal = (tableint *) (data + 1);
-        #ifdef USE_SSE
+#ifdef USE_SSE
                 _mm_prefetch((char *) (visited_array + *(data + 1)), _MM_HINT_T0);
                 _mm_prefetch((char *) (visited_array + *(data + 1) + 64), _MM_HINT_T0);
                 _mm_prefetch(getDataByInternalId(*datal), _MM_HINT_T0);
                 _mm_prefetch(getDataByInternalId(*(datal + 1)), _MM_HINT_T0);
-        #endif
+#endif
 
-                for (int j = 0; j < size; j++) {
+                for (size_t j = 0; j < size; j++) {
                     tableint candidate_id = *(datal + j);
-        #ifdef USE_SSE
+//                    if (candidate_id == 0) continue;
+#ifdef USE_SSE
                     _mm_prefetch((char *) (visited_array + *(datal + j + 1)), _MM_HINT_T0);
                     _mm_prefetch(getDataByInternalId(*(datal + j + 1)), _MM_HINT_T0);
-        #endif
+#endif
                     if (visited_array[candidate_id] == visited_array_tag) continue;
                     visited_array[candidate_id] = visited_array_tag;
                     char *currObj1 = (getDataByInternalId(candidate_id));
 
                     dist_t dist1 = fstdistfunc_(data_point, currObj1, dist_func_param_);
-                    if (top_candidates.top().first > dist1 || top_candidates.size() < ef_construction_) {
+                    if (top_candidates.size() < ef_construction_ || lowerBound > dist1) {
                         candidateSet.emplace(-dist1, candidate_id);
-        #ifdef USE_SSE
+#ifdef USE_SSE
                         _mm_prefetch(getDataByInternalId(candidateSet.top().second), _MM_HINT_T0);
-        #endif
-                        top_candidates.emplace(dist1, candidate_id);
-                        if (top_candidates.size() > ef_construction_) {
+#endif
+
+                        if (!isMarkedDeleted(candidate_id))
+                            top_candidates.emplace(dist1, candidate_id);
+
+                        if (top_candidates.size() > ef_construction_)
                             top_candidates.pop();
-                        }
-                        lowerBound = top_candidates.top().first;
+
+                        if (!top_candidates.empty())
+                            lowerBound = top_candidates.top().first;
                     }
                 }
             }
@@ -229,6 +240,7 @@ namespace hnswlib {
             return top_candidates;
         }
 
+        template <bool has_deletions>
         std::priority_queue<std::pair<dist_t, tableint>, std::vector<std::pair<dist_t, tableint>>, CompareByFirst>
         searchBaseLayerST(tableint ep_id, const void *data_point, size_t ef) const {
             VisitedList *vl = visited_list_pool_->getFreeVisitedList();
@@ -237,39 +249,49 @@ namespace hnswlib {
 
             std::priority_queue<std::pair<dist_t, tableint>, std::vector<std::pair<dist_t, tableint>>, CompareByFirst> top_candidates;
             std::priority_queue<std::pair<dist_t, tableint>, std::vector<std::pair<dist_t, tableint>>, CompareByFirst> candidate_set;
-            dist_t dist = fstdistfunc_(data_point, getDataByInternalId(ep_id), dist_func_param_);
 
-            top_candidates.emplace(dist, ep_id);
-            candidate_set.emplace(-dist, ep_id);
+            dist_t lowerBound;
+            if (!has_deletions || !isMarkedDeleted(ep_id)) {
+                dist_t dist = fstdistfunc_(data_point, getDataByInternalId(ep_id), dist_func_param_);
+                lowerBound = dist;
+                top_candidates.emplace(dist, ep_id);
+                candidate_set.emplace(-dist, ep_id);
+            } else {
+                lowerBound = std::numeric_limits<dist_t>::max();
+                candidate_set.emplace(-lowerBound, ep_id);
+            }
+
             visited_array[ep_id] = visited_array_tag;
-            dist_t lower_bound = dist;
 
             while (!candidate_set.empty()) {
 
                 std::pair<dist_t, tableint> current_node_pair = candidate_set.top();
 
-                if ((-current_node_pair.first) > lower_bound) {
+                if ((-current_node_pair.first) > lowerBound) {
                     break;
                 }
                 candidate_set.pop();
 
                 tableint current_node_id = current_node_pair.second;
-                int *data = (int *) (data_level0_memory_ + current_node_id * size_data_per_element_ + offsetLevel0_);
-                int size = *data;
-        #ifdef USE_SSE
+                int *data = (int *) get_linklist0(current_node_id);
+                size_t size = getListCount((linklistsizeint*)data);
+//                bool cur_node_deleted = isMarkedDeleted(current_node_id);
+
+#ifdef USE_SSE
                 _mm_prefetch((char *) (visited_array + *(data + 1)), _MM_HINT_T0);
                 _mm_prefetch((char *) (visited_array + *(data + 1) + 64), _MM_HINT_T0);
                 _mm_prefetch(data_level0_memory_ + (*(data + 1)) * size_data_per_element_ + offsetData_, _MM_HINT_T0);
                 _mm_prefetch((char *) (data + 2), _MM_HINT_T0);
-        #endif
+#endif
 
-                for (int j = 1; j <= size; j++) {
+                for (size_t j = 1; j <= size; j++) {
                     int candidate_id = *(data + j);
-        #ifdef USE_SSE
+//                    if (candidate_id == 0) continue;
+#ifdef USE_SSE
                     _mm_prefetch((char *) (visited_array + *(data + j + 1)), _MM_HINT_T0);
                     _mm_prefetch(data_level0_memory_ + (*(data + j + 1)) * size_data_per_element_ + offsetData_,
                                  _MM_HINT_T0);////////////
-        #endif
+#endif
                     if (!(visited_array[candidate_id] == visited_array_tag)) {
 
                         visited_array[candidate_id] = visited_array_tag;
@@ -277,20 +299,22 @@ namespace hnswlib {
                         char *currObj1 = (getDataByInternalId(candidate_id));
                         dist_t dist = fstdistfunc_(data_point, currObj1, dist_func_param_);
 
-                        if (top_candidates.top().first > dist || top_candidates.size() < ef) {
+                        if (top_candidates.size() < ef || lowerBound > dist) {
                             candidate_set.emplace(-dist, candidate_id);
-        #ifdef USE_SSE
+#ifdef USE_SSE
                             _mm_prefetch(data_level0_memory_ + candidate_set.top().second * size_data_per_element_ +
                                          offsetLevel0_,///////////
                                          _MM_HINT_T0);////////////////////////
-        #endif
+#endif
 
-                            top_candidates.emplace(dist, candidate_id);
+                            if (!has_deletions || !isMarkedDeleted(candidate_id))
+                                top_candidates.emplace(dist, candidate_id);
 
-                            if (top_candidates.size() > ef) {
+                            if (top_candidates.size() > ef)
                                 top_candidates.pop();
-                            }
-                            lower_bound = top_candidates.top().first;
+
+                            if (!top_candidates.empty())
+                                lowerBound = top_candidates.top().first;
                         }
                     }
                 }
@@ -344,15 +368,15 @@ namespace hnswlib {
         }
 
 
-        linklistsizeint *get_linklist0(tableint internal_id) {
+        linklistsizeint *get_linklist0(tableint internal_id) const {
             return (linklistsizeint *) (data_level0_memory_ + internal_id * size_data_per_element_ + offsetLevel0_);
         };
 
-        linklistsizeint *get_linklist0(tableint internal_id, char *data_level0_memory_) {
+        linklistsizeint *get_linklist0(tableint internal_id, char *data_level0_memory_) const {
             return (linklistsizeint *) (data_level0_memory_ + internal_id * size_data_per_element_ + offsetLevel0_);
         };
 
-        linklistsizeint *get_linklist(tableint internal_id, int level) {
+        linklistsizeint *get_linklist(tableint internal_id, int level) const {
             return (linklistsizeint *) (linkLists_[internal_id] + (level - 1) * size_links_per_element_);
         };
 
@@ -371,6 +395,7 @@ namespace hnswlib {
                 selectedNeighbors.push_back(top_candidates.top().second);
                 top_candidates.pop();
             }
+
             {
                 linklistsizeint *ll_cur;
                 if (level == 0)
@@ -381,7 +406,7 @@ namespace hnswlib {
                 if (*ll_cur) {
                     throw std::runtime_error("The newly inserted element should have blank link list");
                 }
-                *ll_cur = selectedNeighbors.size();
+                setListCount(ll_cur,selectedNeighbors.size());
                 tableint *data = (tableint *) (ll_cur + 1);
 
 
@@ -405,8 +430,8 @@ namespace hnswlib {
                     ll_other = get_linklist0(selectedNeighbors[idx]);
                 else
                     ll_other = get_linklist(selectedNeighbors[idx], level);
-                size_t sz_link_list_other = *ll_other;
 
+                size_t sz_link_list_other = getListCount(ll_other);
 
                 if (sz_link_list_other > Mcurmax)
                     throw std::runtime_error("Bad value of sz_link_list_other");
@@ -418,7 +443,7 @@ namespace hnswlib {
                 tableint *data = (tableint *) (ll_other + 1);
                 if (sz_link_list_other < Mcurmax) {
                     data[sz_link_list_other] = cur_c;
-                    *ll_other = sz_link_list_other + 1;
+                    setListCount(ll_other, sz_link_list_other + 1);
                 } else {
                     // finding the "weakest" element to replace it with the new one
                     dist_t d_max = fstdistfunc_(getDataByInternalId(cur_c), getDataByInternalId(selectedNeighbors[idx]),
@@ -441,7 +466,7 @@ namespace hnswlib {
                         candidates.pop();
                         indx++;
                     }
-                    *ll_other = indx;
+                    setListCount(ll_other, indx);
                     // Nearest K:
                     /*int indx = -1;
                     for (int j = 0; j < sz_link_list_other; j++) {
@@ -476,8 +501,8 @@ namespace hnswlib {
                 while (changed) {
                     changed = false;
                     int *data;
-                    data = (int *) (linkLists_[currObj] + (level - 1) * size_links_per_element_);
-                    int size = *data;
+                    data = (int *) get_linklist(currObj,level);
+                    int size = getListCount(data);
                     tableint *datal = (tableint *) (data + 1);
                     for (int i = 0; i < size; i++) {
                         tableint cand = datal[i];
@@ -494,15 +519,56 @@ namespace hnswlib {
                 }
             }
 
-            //std::priority_queue< std::pair< dist_t, tableint  >> top_candidates = searchBaseLayer(currObj, query_data, 0);
-            std::priority_queue<std::pair<dist_t, tableint  >> top_candidates = searchBaseLayerST(currObj, query_data,
-                                                                                                  ef_);
+
+            std::priority_queue<std::pair<dist_t, tableint  >> top_candidates;
+            if (has_deletions_) {
+                std::priority_queue<std::pair<dist_t, tableint  >> top_candidates1=searchBaseLayerST<true>(currObj, query_data,
+                                                                                                             ef_);
+                top_candidates.swap(top_candidates1);
+            }
+            else{
+                std::priority_queue<std::pair<dist_t, tableint  >> top_candidates1=searchBaseLayerST<false>(currObj, query_data,
+                                                                                                              ef_);
+                top_candidates.swap(top_candidates1);
+            }
+
             while (top_candidates.size() > k) {
                 top_candidates.pop();
             }
             return top_candidates;
         };
 
+        void resizeIndex(size_t new_max_elements){
+            if (new_max_elements<cur_element_count)
+                throw std::runtime_error("Cannot resize, max element is less than the current number of elements");
+
+
+            delete visited_list_pool_;
+            visited_list_pool_ = new VisitedListPool(1, new_max_elements);
+
+
+
+            element_levels_.resize(new_max_elements);
+
+            std::vector<std::mutex>(new_max_elements).swap(link_list_locks_);
+
+
+            // Reallocate base layer
+            char * data_level0_memory_new = (char *) malloc(new_max_elements * size_data_per_element_);
+            memcpy(data_level0_memory_new, data_level0_memory_,cur_element_count * size_data_per_element_);
+            free(data_level0_memory_);
+            data_level0_memory_=data_level0_memory_new;
+
+            // Reallocate all other layers
+            char ** linkLists_new = (char **) malloc(sizeof(void *) * new_max_elements);
+            memcpy(linkLists_new, linkLists_,cur_element_count * sizeof(void *));
+            free(linkLists_);
+            linkLists_=linkLists_new;
+
+            max_elements_=new_max_elements;
+
+        }
+
         void saveIndex(const std::string &location) {
             std::ofstream output(location, std::ios::binary);
             std::streampos position;
@@ -538,6 +604,10 @@ namespace hnswlib {
 
             std::ifstream input(location, std::ios::binary);
 
+            if (!input.is_open())
+                throw std::runtime_error("Cannot open file");
+
+
             // get file size:
             input.seekg(0,input.end);
             std::streampos total_filesize=input.tellg();
@@ -568,16 +638,15 @@ namespace hnswlib {
             fstdistfunc_ = s->get_dist_func();
             dist_func_param_ = s->get_dist_func_param();
 
-            /// Legacy, check that everything is ok
-
-            bool old_index=false;
-
             auto pos=input.tellg();
+
+
+            /// Optional - check if index is ok:
+
             input.seekg(cur_element_count * size_data_per_element_,input.cur);
             for (size_t i = 0; i < cur_element_count; i++) {
                 if(input.tellg() < 0 || input.tellg()>=total_filesize){
-                    old_index = true;
-                    break;
+                    throw std::runtime_error("Index seems to be corrupted or unsupported");
                 }
 
                 unsigned int linkListSize;
@@ -587,23 +656,21 @@ namespace hnswlib {
                 }
             }
 
-            // check if file is ok, if not this is either corrupted or old index
+            // throw exception if it either corrupted or old index
             if(input.tellg()!=total_filesize)
-                old_index = true;
+                throw std::runtime_error("Index seems to be corrupted or unsupported");
 
-            if (old_index) {
-                showUpdate("Warning: loading of old indexes will be deprecated before 2019.\n"
-                           "Please resave the index in the new format.\n");
-            }
             input.clear();
+
+            /// Optional check end
+
             input.seekg(pos,input.beg);
 
 
             data_level0_memory_ = (char *) malloc(max_elements * size_data_per_element_);
             input.read(data_level0_memory_, cur_element_count * size_data_per_element_);
 
-            if(old_index)
-                input.seekg(((max_elements_-cur_element_count) * size_data_per_element_), input.cur);
+
 
 
             size_links_per_element_ = maxM_ * sizeof(tableint) + sizeof(linklistsizeint);
@@ -634,6 +701,14 @@ namespace hnswlib {
                     input.read(linkLists_[i], linkListSize);
                 }
             }
+
+            has_deletions_=false;
+
+            for (size_t i = 0; i < cur_element_count; i++) {
+                if(isMarkedDeleted(i))
+                    has_deletions_=true;
+            }
+
             input.close();
 
             return;
@@ -642,41 +717,100 @@ namespace hnswlib {
         template<typename data_t>
         std::vector<data_t> getDataByLabel(labeltype label)
         {
-          tableint label_c;
-          auto search = label_lookup_.find(label);
-          if (search == label_lookup_.end()) {
-            throw std::runtime_error("Label not found");
-          }
-          label_c = search->second;
+            tableint label_c;
+            auto search = label_lookup_.find(label);
+            if (search == label_lookup_.end() || isMarkedDeleted(search->second)) {
+                throw std::runtime_error("Label not found");
+            }
+            label_c = search->second;
 
-          char* data_ptrv = getDataByInternalId(label_c);
-          size_t dim = *((size_t *) dist_func_param_);
-          std::vector<data_t> data;
-          data_t* data_ptr = (data_t*) data_ptrv;
-          for (int i = 0; i < dim; i++) {
-            data.push_back(*data_ptr);
-            data_ptr += 1;
-          }
-          return data;
+            char* data_ptrv = getDataByInternalId(label_c);
+            size_t dim = *((size_t *) dist_func_param_);
+            std::vector<data_t> data;
+            data_t* data_ptr = (data_t*) data_ptrv;
+            for (int i = 0; i < dim; i++) {
+                data.push_back(*data_ptr);
+                data_ptr += 1;
+            }
+            return data;
         }
 
-        void addPoint(void *data_point, labeltype label)
+        static const unsigned char DELETE_MARK = 0x01;
+//        static const unsigned char REUSE_MARK = 0x10;
+        /**
+         * Marks an element with the given label deleted, does NOT really change the current graph.
+         * @param label
+         */
+        void markDelete(labeltype label)
         {
+            has_deletions_=true;
+            auto search = label_lookup_.find(label);
+            if (search == label_lookup_.end()) {
+                throw std::runtime_error("Label not found");
+            }
+            markDeletedInternal(search->second);
+        }
+
+        /**
+         * Uses the first 8 bits of the memory for the linked list to store the mark,
+         * whereas maxM0_ has to be limited to the lower 24 bits, however, still large enough in almost all cases.
+         * @param internalId
+         */
+        void markDeletedInternal(tableint internalId) {
+            unsigned char *ll_cur = ((unsigned char *)get_linklist0(internalId))+2;
+            *ll_cur |= DELETE_MARK;
+        }
+
+        /**
+         * Remove the deleted mark of the node.
+         * @param internalId
+         */
+        void unmarkDeletedInternal(tableint internalId) {
+            unsigned char *ll_cur = ((unsigned char *)get_linklist0(internalId))+2;
+            *ll_cur &= ~DELETE_MARK;
+        }
+
+        /**
+         * Checks the first 8 bits of the memory to see if the element is marked deleted.
+         * @param internalId
+         * @return
+         */
+        bool isMarkedDeleted(tableint internalId) const {
+            unsigned char *ll_cur = ((unsigned char*)get_linklist0(internalId))+2;
+            return *ll_cur & DELETE_MARK;
+        }
+
+        unsigned short int getListCount(linklistsizeint * ptr) const {
+            return *((unsigned short int *)ptr);
+        }
+
+        void setListCount(linklistsizeint * ptr, unsigned short int size) const {
+            *((unsigned short int*)(ptr))=*((unsigned short int *)&size);
+        }
+
+        void addPoint(void *data_point, labeltype label) {
             addPoint(data_point, label,-1);
         }
 
         tableint addPoint(void *data_point, labeltype label, int level) {
-
             tableint cur_c = 0;
             {
                 std::unique_lock <std::mutex> lock(cur_element_count_guard_);
                 if (cur_element_count >= max_elements_) {
                     throw std::runtime_error("The number of elements exceeds the specified limit");
                 };
+
                 cur_c = cur_element_count;
-                label_lookup_[label] = cur_c;  // expected unique, if not will overwrite
                 cur_element_count++;
+
+                auto search = label_lookup_.find(label);
+                if (search != label_lookup_.end()) {
+                    std::unique_lock <std::mutex> lock_el(link_list_locks_[search->second]);
+                    markDeletedInternal(search->second);
+                }
+                label_lookup_[label] = cur_c;
             }
+
             std::unique_lock <std::mutex> lock_el(link_list_locks_[cur_c]);
             int curlevel = getRandomLevel(mult_);
             if (level > 0)
@@ -690,6 +824,7 @@ namespace hnswlib {
             if (curlevel <= maxlevelcopy)
                 templock.unlock();
             tableint currObj = enterpoint_node_;
+            tableint enterpoint_copy = enterpoint_node_;
 
 
             memset(data_level0_memory_ + cur_c * size_data_per_element_ + offsetLevel0_, 0, size_data_per_element_);
@@ -703,8 +838,8 @@ namespace hnswlib {
                 linkLists_[cur_c] = (char *) malloc(size_links_per_element_ * curlevel + 1);
                 memset(linkLists_[cur_c], 0, size_links_per_element_ * curlevel + 1);
             }
-            if ((signed)currObj != -1) {
 
+            if ((signed)currObj != -1) {
 
                 if (curlevel < maxlevelcopy) {
 
@@ -715,10 +850,11 @@ namespace hnswlib {
                         bool changed = true;
                         while (changed) {
                             changed = false;
-                            int *data;
+                            unsigned int *data;
                             std::unique_lock <std::mutex> lock(link_list_locks_[currObj]);
-                            data = (int *) (linkLists_[currObj] + (level - 1) * size_links_per_element_);
-                            int size = *data;
+                            data = get_linklist(currObj,level);
+                            int size = getListCount(data);
+
                             tableint *datal = (tableint *) (data + 1);
                             for (int i = 0; i < size; i++) {
                                 tableint cand = datal[i];
@@ -735,13 +871,21 @@ namespace hnswlib {
                     }
                 }
 
+                bool epDeleted = isMarkedDeleted(enterpoint_copy);
                 for (int level = std::min(curlevel, maxlevelcopy); level >= 0; level--) {
-                    if (level > maxlevelcopy || level < 0)
+                    if (level > maxlevelcopy || level < 0)  // possible?
                         throw std::runtime_error("Level error");
 
                     std::priority_queue<std::pair<dist_t, tableint>, std::vector<std::pair<dist_t, tableint>>, CompareByFirst> top_candidates = searchBaseLayer(
                             currObj, data_point, level);
+                    if (epDeleted) {
+                        top_candidates.emplace(fstdistfunc_(data_point, getDataByInternalId(enterpoint_copy), dist_func_param_), enterpoint_copy);
+                        if (top_candidates.size() > ef_construction_)
+                            top_candidates.pop();
+                    }
                     mutuallyConnectNewElement(data_point, cur_c, top_candidates, level);
+
+                    currObj = top_candidates.top().second;
                 }
 
 
@@ -768,9 +912,10 @@ namespace hnswlib {
                 bool changed = true;
                 while (changed) {
                     changed = false;
-                    int *data;
-                    data = (int *) (linkLists_[currObj] + (level - 1) * size_links_per_element_);
-                    int size = *data;
+                    unsigned int *data;
+
+                    data = (unsigned int *) get_linklist(currObj, level);
+                    int size = getListCount(data);
                     tableint *datal = (tableint *) (data + 1);
                     for (int i = 0; i < size; i++) {
                         tableint cand = datal[i];
@@ -787,9 +932,17 @@ namespace hnswlib {
                 }
             }
 
-
-            std::priority_queue<std::pair<dist_t, tableint>, std::vector<std::pair<dist_t, tableint>>, CompareByFirst> top_candidates = searchBaseLayerST(
-                    currObj, query_data, std::max(ef_,k));
+            std::priority_queue<std::pair<dist_t, tableint>, std::vector<std::pair<dist_t, tableint>>, CompareByFirst> top_candidates;
+            if (has_deletions_) {
+                std::priority_queue<std::pair<dist_t, tableint>, std::vector<std::pair<dist_t, tableint>>, CompareByFirst> top_candidates1=searchBaseLayerST<true>(
+                        currObj, query_data, std::max(ef_, k));
+                top_candidates.swap(top_candidates1);
+            }
+            else{
+                std::priority_queue<std::pair<dist_t, tableint>, std::vector<std::pair<dist_t, tableint>>, CompareByFirst> top_candidates1=searchBaseLayerST<false>(
+                        currObj, query_data, std::max(ef_, k));
+                top_candidates.swap(top_candidates1);
+            }
             std::priority_queue<std::pair<dist_t, labeltype >> results;
             while (top_candidates.size() > k) {
                 top_candidates.pop();
